@@ -23,52 +23,53 @@ import re
 import sys
 
 # External modules
-from configobj import ConfigObj
+from ConfigParser import RawConfigParser
+
 try:
     # Voluptuous version >= 0.8.1
     from voluptuous import message, MultipleInvalid, Invalid, Schema
-    from voluptuous import All, Length, Any, Url, Boolean
+    from voluptuous import All, Length, Any, Url, Boolean, Optional
 except ImportError:
     # Voluptuous version <= 0.7.2
     from voluptuous.voluptuous import message, MultipleInvalid, Invalid
-    from voluptuous import Schema, All, Length, Any, Url, Boolean
+    from voluptuous import Schema, All, Length, Any, Url, Boolean, Optional
 
 # Elasticluster imports
+from elasticluster import log
 from elasticluster.exceptions import ConfigurationError
-from elasticluster.providers.ec2_boto import BotoCloudProvider
-from elasticluster.providers.gce import GoogleCloudProvider
 from elasticluster.providers.ansible_provider import AnsibleSetupProvider
-from elasticluster.cluster import Cluster, ClusterStorage, Node
+from elasticluster.cluster import Cluster
+from elasticluster.repository import ClusterRepository
 
 
 class Configurator(object):
-    """
-    The Configurator is responsible for (I) keeping track of the
+    """The Configurator is responsible for (I) keeping track of the
     configuration and (II) offer factory methods to create all kind of
     objects that need information from the configuration.
 
     The cluster configuration dictionary is structured in the following way:
     (see an example @
      https://github.com/gc3-uzh-ch/elasticluster/wiki/Configuration-Module)
-    { "<cluster_template>" : {
-        "setup" : { properties of the setup section },
-        "cloud" : { properties of the cloud section },
-        "login" : { properties of the login section },
-        "cluster" : { properties of the cluster section },
-        "nodes": {  "<node_type>" : { properties of the node},
-                    "<node_type>" : { properties of the node},
-                },
-        },
-     "<cluster_template>" : {
-        (see above)
-        }
-     }
-    """
+    ::
 
-    cloud_providers_map = {
-        "ec2_boto": BotoCloudProvider,
-        "google": GoogleCloudProvider,
-    }
+        { "<cluster_template>" : {
+            "setup" : { properties of the setup section },
+            "cloud" : { properties of the cloud section },
+            "login" : { properties of the login section },
+            "cluster" : { properties of the cluster section },
+            "nodes": {  "<node_kind>" : { properties of the node},
+                        "<node_kind>" : { properties of the node},
+                    },
+            },
+         "<cluster_template>" : {
+            (see above)
+            }
+         }
+
+   :param dict cluster_conf: see description above
+   :param str storage_path: path to store data
+   :raises MultipleInvalid: configuration validation
+    """
 
     setup_providers_map = {"ansible": AnsibleSetupProvider, }
 
@@ -76,15 +77,12 @@ class Configurator(object):
         "~/.elasticluster/storage")
 
     def __init__(self, cluster_conf, storage_path=None):
-        """
-        Default constructor to initialize a Configurator.
-        :param cluster_conf: configuration dictionary
-        :raises MultipleInvalid: configuration validation
-        """
         self.general_conf = dict()
         self.cluster_conf = cluster_conf
 
         if storage_path:
+            storage_path = os.path.expanduser(storage_path)
+            storage_path = os.path.expandvars(storage_path)
             self.general_conf['storage'] = storage_path
         else:
             self.general_conf['storage'] = Configurator.default_storage_dir
@@ -94,27 +92,41 @@ class Configurator(object):
 
     @classmethod
     def fromConfig(cls, configfile, storage_path=None):
-        """
-        Helper method to initialize Configurator from an ini file.
-        :param configfile: path to the ini file
-        :return: configurator object
+        """Helper method to initialize Configurator from an ini file.
+
+        :param str configfile: path to the ini file
+        :return: :py:class:`Configurator`
         """
         config_reader = ConfigReader(configfile)
         conf = config_reader.read_config()
         return Configurator(conf, storage_path=storage_path)
 
     def create_cloud_provider(self, cluster_template):
-        """
-        Creates a cloud provider by inspecting the configuration properties
+        """Creates a cloud provider by inspecting the configuration properties
         of the given cluster template.
-        :param cluster_template: template to use (if not already specified
-            on init)
-        :return: object that fulfills the contract of
-            :py:class:`elasticluster.providers.AbstractSetupProvider`
+
+        :param str cluster_template: template to use (if not already specified
+                                 on init)
+        :return: cloud provider that fulfills the contract of
+                 :py:class:`elasticluster.providers.AbstractSetupProvider`
         """
         conf = self.cluster_conf[cluster_template]['cloud']
 
-        provider = Configurator.cloud_providers_map[conf['provider']]
+        try:
+            if conf['provider'] == 'ec2_boto':
+                from elasticluster.providers.ec2_boto import BotoCloudProvider
+                provider = BotoCloudProvider
+            elif conf['provider'] == 'openstack':
+                from elasticluster.providers.openstack import OpenStackCloudProvider
+                provider = OpenStackCloudProvider
+            elif conf['provider'] == 'google':
+                from elasticluster.providers.gce import GoogleCloudProvider
+                provider = GoogleCloudProvider
+            else:
+                raise Invalid("Invalid provider '%s' for cluster '%s'"% (conf['provider'], cluster_template))
+        except ImportError, ex:
+            raise Invalid("Unable to load provider '%s': %s" % (conf['provider'], ex))
+
         providerconf = conf.copy()
         providerconf.pop('provider')
         providerconf['storage_path'] = self.general_conf['storage']
@@ -122,12 +134,12 @@ class Configurator(object):
         return provider(**providerconf)
 
     def create_cluster(self, template, name=None):
-        """
-        Creates a cluster by inspecting the configuration properties of the
-            given cluster template.
-        :param template: name of the cluster template
+        """Creates a cluster by inspecting the configuration properties of the
+        given cluster template.
 
-        :param name: name of the cluster. If not defined, the cluster
+        :param str template: name of the cluster template
+
+        :param str name: name of the cluster. If not defined, the cluster
         will be named after the template.
 
         :return: :py:class:`elasticluster.cluster.cluster` instance
@@ -143,78 +155,62 @@ class Configurator(object):
                 "" % (template, name))
 
         conf = self.cluster_conf[template]
-
-        nodes = dict(
-            (k[:-6], int(v)) for k, v in conf['cluster'].iteritems() if
-            k.endswith('_nodes'))
-        min_nodes = dict(
-            (k[:-10], int(v)) for k, v in conf['cluster'].iteritems() if
-            k.endswith('_nodes_min'))
+        conf_login = self.cluster_conf[template]['login']
 
         extra = conf['cluster'].copy()
         extra.pop('cloud')
         extra.pop('setup_provider')
-        return Cluster(template,
-                       name,
-                       conf['cluster']['cloud'],
-                       self.create_cloud_provider(template),
-                       self.create_setup_provider(template, name=name),
-                       nodes,
-                       self, min_nodes=min_nodes, **extra)
+        extra['template'] = template
 
-    def load_cluster(self, cluster_name):
-        """
-        Loads a cluster from the local stored information.
-        :param cluster_name: name of the cluster
-        :return: :py:class:`elasticluster.cluster.cluster` instance
-        """
-        storage = self.create_cluster_storage()
-        information = storage.load_cluster(cluster_name)
+        cluster = Cluster(name,
+                          self.create_cloud_provider(template),
+                          self.create_setup_provider(template, name=name),
+                          conf_login['user_key_name'],
+                          conf_login['user_key_public'],
+                          conf_login["user_key_private"],
+                          repository=self.create_repository(),
+                          **extra)
 
-        cluster = self.create_cluster(
-            information['template'], information['name'])
-
-        # Clear cluster nodes.
-        cluster.nodes = dict((k, []) for k in cluster.nodes)
-        for dnode in information['nodes']:
-            if dnode['type'] not in cluster.nodes:
-                cluster.nodes[dnode['type']] = []
-            node = cluster.add_node(dnode['type'], name=dnode['name'])
-            node.instance_id = dnode['instance_id']
-            node.ip_public = dnode['ip_public']
-            node.ip_private = dnode['ip_private']
+        nodes = dict(
+            (k[:-6], int(v)) for k, v in conf['cluster'].iteritems() if
+            k.endswith('_nodes'))
 
+        for kind, num in nodes.iteritems():
+            conf_kind = conf['nodes'][kind]
+            extra = conf_kind.copy()
+            extra.pop('image_id', None)
+            extra.pop('flavor', None)
+            extra.pop('security_group', None)
+            extra.pop('image_userdata', None)
+            userdata = conf_kind.get('image_userdata', '')
+            cluster.add_nodes(kind,
+                              num,
+                              conf_kind['image_id'],
+                              conf_login['image_user'],
+                              conf_kind['flavor'],
+                              conf_kind['security_group'],
+                              image_userdata=userdata,
+                              **extra)
         return cluster
 
-    def create_node(self, cluster_template, node_type, cloud_provider, name):
-        """
-        :param cluster_template: name of the cluster template
-        :param node_type: type of the node, string defining the <group> in
-            the configuration
-        :param cloud_provider: instance of :py:class:`elasticluster
-            .providers.AbstractCloudProvider`
-        :param name: name of the node
-        :return: :py:class:`elasticluster.cluster.node` instance
-        """
-        conf = self.cluster_conf[cluster_template]['nodes'][node_type]
-        conf_login = self.cluster_conf[cluster_template]['login']
+    def load_cluster(self, cluster_name):
+        """Loads a cluster from the cluster repository.
 
-        return Node(name, node_type, cloud_provider,
-                    conf_login['user_key_public'],
-                    conf_login["user_key_private"],
-                    conf_login['user_key_name'],
-                    conf_login['image_user'], conf['security_group'],
-                    conf['image_id'], conf['flavor'],
-                    image_userdata=conf.get('image_userdata', ''))
-
-    def create_cluster_storage(self):
+        :param str cluster_name: name of the cluster
+        :return: :py:class:`elasticluster.cluster.cluster` instance
         """
-        Creates an instance of :py:class:`elasticluster.cluster.ClusterStorage`
-            to safe information about a cluster local.
-        """
-        return ClusterStorage(self.general_conf['storage'])
+        repository = self.create_repository()
+        cluster = repository.get(cluster_name)
+        if not cluster._setup_provider:
+            cluster._setup_provider = self.create_setup_provider(cluster.extra['template'])
+        return cluster
 
     def create_setup_provider(self, cluster_template, name=None):
+        """Creates the setup provider for the given cluster template.
+
+        :param str cluster_template: template of the cluster
+        :param str name: name of the cluster to read configuration properties
+        """
         conf = self.cluster_conf[cluster_template]['setup']
         conf['general_conf'] = self.general_conf.copy()
         if name:
@@ -227,30 +223,54 @@ class Configurator(object):
                 "Invalid value `%s` for `setup_provider` in configuration "
                 "file." % provider_name)
 
-        provider = Configurator.setup_providers_map[provider_name]
+        storage_path = self.general_conf['storage']
+        if 'playbook_path' in conf:
+            playbook_path = conf['playbook_path']
+            del conf['playbook_path']
+        else:
+            playbook_path = None
+        groups = dict((k[:-7], v.split(',')) for k, v
+                      in conf.items() if k.endswith('_groups'))
+        environment = dict()
+        for nodekind, grps in groups.iteritems():
+            if not isinstance(grps, list):
+                groups[nodekind] = [grps]
 
-        return provider(
-            conf_login['user_key_private'], conf_login['image_user'],
-            conf_login['image_user_sudo'], conf_login['image_sudo'],
-            **conf)
+            # Environment variables parsing
+            environment[nodekind] = dict()
+            for key, value in conf.iteritems():
+                prefix = "%s_var_" % nodekind
+                if key.startswith(prefix):
+                    var = key.replace(prefix, '')
+                    environment[nodekind][var] = value
+
+        provider = Configurator.setup_providers_map[provider_name]
+        return provider(groups, playbook_path=playbook_path,
+                        environment_vars=environment,
+                        storage_path=storage_path,
+                        sudo_user=conf_login['image_user_sudo'],
+                        sudo=conf_login['image_sudo'], **conf)
+
+    def create_repository(self):
+        storage_path = self.general_conf['storage']
+        repository = ClusterRepository(storage_path)
+        return repository
 
 
 class ConfigValidator(object):
-    """
-    Validator for the cluster configuration dictionary.
+    """Validator for the cluster configuration dictionary.
+
+    :param config: dictionary containing cluster configuration properties
     """
 
     def __init__(self, config):
-        """
-        :param config: dictionary containing cluster configuration properties
-        """
         self.config = config
 
     def _pre_validate(self):
-        """
-        Handles all pre validation phase functionality, such as:
-        - reading environment variables
-        - interpolating configuraiton options
+        """Handles all pre validation phase functionality, such as:
+
+        * reading environment variables
+        * interpolating configuraiton options
         """
         # read cloud provider environment variables (ec2_boto or google)
         for cluster, props in self.config.iteritems():
@@ -274,22 +294,17 @@ class ConfigValidator(object):
                                             str(ansible_pb_dir))
                     self.config[cluster]['setup']['playbook_path'] = pbpath
 
-            elif 'setup' in props:
-                # set default playbook directory if none specified
-                setup_conf = self.config[cluster]['setup']
-                setup_conf['playbook_path'] = ansible_pb_dir + os.sep + \
-                                              "site.yml"
-
     def _post_validate(self):
-        """
-        Handles all post validation phase functionality, such as:
-        - expanding file paths
+        """Handles all post validation phase functionality, such as:
+
+        * expanding file paths
         """
         # expand all paths
         for cluster, values in self.config.iteritems():
             conf = self.config[cluster]
-            pbpath = os.path.expanduser(values['setup']['playbook_path'])
-            conf['setup']['playbook_path'] = pbpath
+            if 'playbook_path' in values['setup']:
+                pbpath = os.path.expanduser(values['setup']['playbook_path'])
+                conf['setup']['playbook_path'] = pbpath
 
             privkey = os.path.expanduser(values['login']['user_key_private'])
             conf['login']['user_key_private'] = privkey
@@ -298,10 +313,13 @@ class ConfigValidator(object):
             conf['login']['user_key_public'] = pubkey
 
     def validate(self):
-        """
-        Validates the given configuration :py:attr:`self.config` to comply
+        """Validates the given configuration :py:attr:`self.config` to comply
         with elasticluster. As well all types are converted to the expected
         format if possible.
+
+        :raisegs: :py:class:`voluptuous.MultipleInvalid` if multiple
+                 properties are not compliant
+        :raises: :py:class:`voluptuous.Invalid` if one property is invalid
         """
         self._pre_validate()
         # custom validators
@@ -313,37 +331,53 @@ class ConfigValidator(object):
             else:
                 raise Invalid("file could not be found `%s`" % v)
 
+        @message("Unsupported nova API version")
+        def nova_api_version(version):
+            try:
+                from novaclient import client,exceptions
+                client.get_client_class(version)
+                return version
+            except exceptions.UnsupportedVersion as ex:
+                raise Invalid(
+                    "Invalid option for `nova_api_version`: %s" % ex)
+
         # schema to validate all cluster properties
         schema = {"cluster": {"cloud": All(str, Length(min=1)),
                               "setup_provider": All(str, Length(min=1)),
-                              "login": All(str, Length(min=1)),
-                              },
+                              "login": All(str, Length(min=1))},
                   "setup": {"provider": All(str, Length(min=1)),
-                            "playbook_path": check_file(),
-                            },
+                            Optional("playbook_path"): check_file()},
                   "login": {"image_user": All(str, Length(min=1)),
                             "image_user_sudo": All(str, Length(min=1)),
                             "image_sudo": Boolean(str),
                             "user_key_name": All(str, Length(min=1)),
                             "user_key_private": check_file(),
-                            "user_key_public": check_file(),
-                            }
-                  }
+                            "user_key_public": check_file()}}
 
         cloud_schema_ec2 = {"provider": 'ec2_boto',
                             "ec2_url": Url(str),
                             "ec2_access_key": All(str, Length(min=1)),
                             "ec2_secret_key": All(str, Length(min=1)),
-                            "ec2_region": All(str, Length(min=1))}
+                            "ec2_region": All(str, Length(min=1)),
+                            Optional("request_floating_ip"): Boolean(str)}
         cloud_schema_gce = {"provider": 'google',
                             "gce_client_id": All(str, Length(min=1)),
                             "gce_client_secret": All(str, Length(min=1)),
                             "gce_project_id": All(str, Length(min=1))}
 
+        cloud_schema_openstack = {"provider": 'openstack',
+                                  "auth_url": All(str, Length(min=1)),
+                                  "username": All(str, Length(min=1)),
+                                  "password": All(str, Length(min=1)),
+                                  "project_name": All(str, Length(min=1)),
+                                  Optional("request_floating_ip"): Boolean(str),
+                                  Optional("region_name"): All(str, Length(min=1)),
+                                  Optional("nova_api_version"): nova_api_version()}
+
         node_schema = {
             "flavor": All(str, Length(min=1)),
             "image_id": All(str, Length(min=1)),
-            "security_group": All(str, Length(min=1))
+            "security_group": All(str, Length(min=1)),
         }
 
         # validation
@@ -351,17 +385,25 @@ class ConfigValidator(object):
         validator_node = Schema(node_schema, required=True, extra=True)
         ec2_validator = Schema(cloud_schema_ec2, required=True, extra=False)
         gce_validator = Schema(cloud_schema_gce, required=True, extra=False)
+        openstack_validator = Schema(cloud_schema_openstack, required=True, extra=False)
 
         if not self.config:
             raise Invalid("No clusters found in configuration.")
 
         for cluster, properties in self.config.iteritems():
-            validator(properties)
+            self.config[cluster] = validator(properties)
 
-            if properties['cloud']['provider'] == "ec2":
-                ec2_validator(properties['cloud'])
+            if 'provider' not in properties['cloud']:
+                raise Invalid(
+                    "Missing `provider` option in cluster `%s`" % cluster)
+
+            cloud_props = properties['cloud']
+            if properties['cloud']['provider'] == "ec2_boto":
+                self.config[cluster]['cloud'] = ec2_validator(cloud_props)
             elif properties['cloud']['provider'] == "google":
-                gce_validator(properties['cloud'])
+                self.config[cluster]['cloud'] = gce_validator(cloud_props)
+            elif properties['cloud']['provider'] == "openstack":
+                self.config[cluster]['cloud'] = openstack_validator(cloud_props)
 
             if 'nodes' not in properties or len(properties['nodes']) == 0:
                 raise Invalid(
@@ -382,8 +424,9 @@ class ConfigValidator(object):
 
 
 class ConfigReader(object):
-    """
-    Reads the configuration properties from a ini file.
+    """Reads the configuration properties from a ini file.
+
+    :param str configfile: path to configfile
     """
     cluster_section = "cluster"
     login_section = "login"
@@ -392,11 +435,15 @@ class ConfigReader(object):
     node_section = "node"
 
     def __init__(self, configfile):
-        """
-        :param configfile: path to configfile
-        """
         self.configfile = configfile
-        self.conf = ConfigObj(self.configfile, interpolation=False)
+
+        configparser = RawConfigParser()
+        config_tmp = configparser.read(self.configfile)
+        self.conf = dict()
+        for section in configparser.sections():
+            self.conf[section] = dict(configparser.items(section))
+
+        #self.conf = ConfigObj(self.configfile, interpolation=False)
 
         @message("file could not be found")
         def check_file(v):
@@ -406,43 +453,56 @@ class ConfigReader(object):
             else:
                 raise Invalid("file could not be found `%s`" % v)
 
+        @message("Unsupported nova API version")
+        def nova_api_version(version):
+            try:
+                from novaclient import client, exceptions
+                client.get_client_class(version)
+                return version
+            except exceptions.UnsupportedVersion as ex:
+                raise Invalid(
+                    "Invalid option for `nova_api_version`: %s" % ex)
+
         self.schemas = {
             "cloud": Schema(
-                {"provider": Any('ec2_boto', 'google'),
-                "ec2_url": Url(str),
-                "ec2_access_key": All(str, Length(min=1)),
-                "ec2_secret_key": All(str, Length(min=1)),
-                "ec2_region": All(str, Length(min=1)),
-                "gce_project_id": All(str, Length(min=1)),
-                "gce_client_id": All(str, Length(min=1)),
-                "gce_client_secret": All(str, Length(min=1)),
-                }),
+                {"provider": Any('ec2_boto', 'google', 'openstack'),
+                 "ec2_url": Url(str),
+                 "ec2_access_key": All(str, Length(min=1)),
+                 "ec2_secret_key": All(str, Length(min=1)),
+                 "ec2_region": All(str, Length(min=1)),
+                 "auth_url": All(str, Length(min=1)),
+                 "username": All(str, Length(min=1)),
+                 "password": All(str, Length(min=1)),
+                 "tenant_name": All(str, Length(min=1)),
+                 Optional("region_name"): All(str, Length(min=1)),
+                 "gce_project_id": All(str, Length(min=1)),
+                 "gce_client_id": All(str, Length(min=1)),
+                 "gce_client_secret": All(str, Length(min=1)),
+                 "nova_client_api": nova_api_version()}, extra=True),
             "cluster": Schema(
                 {"cloud": All(str, Length(min=1)),
                  "setup_provider": All(str, Length(min=1)),
-                 "login": All(str, Length(min=1)),
-                 }, required=True, extra=True),
+                 "login": All(str, Length(min=1))}, required=True, extra=True),
             "setup": Schema(
                 {"provider": All(str, Length(min=1)),
-                 }, required=True, extra=True),
+                    }, required=True, extra=True),
             "login": Schema(
                 {"image_user": All(str, Length(min=1)),
                  "image_user_sudo": All(str, Length(min=1)),
                  "image_sudo": Boolean(str),
                  "user_key_name": All(str, Length(min=1)),
                  "user_key_private": check_file(),
-                 "user_key_public": check_file(),
-                 }, required=True)
-            }
+                 "user_key_public": check_file()}, required=True)
+        }
 
     def read_config(self):
-        """
-        Reads the configuration properties from the ini file and links the
+        """Reads the configuration properties from the ini file and links the
         section to comply with the cluster config dictionary format.
+
         :return: dictionary containing all configuration properties from the
          ini file in compliance to the cluster config format
-        :raises MultipleInvalid: not all sections present or broken links
-            between secitons
+        :raises: :py:class:`voluptuous.MultipleInvalid` if not all sections
+                 present or broken links between secitons
         """
         clusters = dict((key, value) for key, value in self.conf.iteritems() if
                         re.search(ConfigReader.cluster_section + "/(.*)", key)
@@ -451,7 +511,6 @@ class ConfigReader(object):
         conf_values = dict()
 
         errors = MultipleInvalid()
-
         for cluster in clusters:
             name = re.search(ConfigReader.cluster_section + "/(.*)",
                              cluster).groups()[0]
@@ -481,7 +540,9 @@ class ConfigReader(object):
                 values['setup'] = dict(self.conf[setup_name])
                 self.schemas['setup'](values['setup'])
             except KeyError, ex:
-                errors.add("cluster `%s` setup section `%s` does not exists" % (cluster, setup_name))
+                errors.add(
+                    "cluster `%s` setup section `%s` does not exists" % (
+                        cluster, setup_name))
             except MultipleInvalid, ex:
                 for error in ex.errors:
                     errors.add(error)
@@ -490,7 +551,9 @@ class ConfigReader(object):
                 values['login'] = dict(self.conf[login_name])
                 self.schemas['login'](values['login'])
             except KeyError, ex:
-                errors.add("cluster `%s` login section `%s` does not exists" % (cluster, login_name))
+                errors.add(
+                    "cluster `%s` login section `%s` does not exists" % (
+                        cluster, login_name))
             except MultipleInvalid, ex:
                 for error in ex.errors:
                     errors.add(error)
@@ -499,10 +562,12 @@ class ConfigReader(object):
                 values['cloud'] = dict(self.conf[cloud_name])
                 self.schemas['cloud'](values['cloud'])
             except KeyError, ex:
-                errors.add("cluster `%s` cloud section `%s` does not exists" % (cluster, cloud_name))
+                errors.add(
+                    "cluster `%s` cloud section `%s` does not exists" % (
+                        cluster, cloud_name))
             except MultipleInvalid, ex:
                 for error in ex.errors:
-                    errors.add(error)
+                    errors.add(Invalid("section %s: %s" % (cloud_name, error)))
 
             try:
                 # nodes can inherit the properties of cluster or overwrite them
@@ -524,7 +589,13 @@ class ConfigReader(object):
                     else:
                         values['nodes'][node_name] = values['cluster']
 
-                conf_values[name] = values
+                if errors.errors:
+                    for error in errors.errors:
+                        log.warning(
+                            "Ignoring Cluster `%s`: %s" % (name, error))
+                    log.warning("Ignoring cluster `%s`." % name)
+                else:
+                    conf_values[name] = values
             except KeyError, ex:
                 errors.add("Error in section `%s`" % cluster)
 
