@@ -71,7 +71,9 @@ class ElastiCluster(cli.app.CommandLineApp):
 
         # global parameters
         self.add_param('-v', '--verbose', action='count', default=0,
-                       help="Increase verbosity.")
+                       help="Increase verbosity. If at least four `-v` option "
+                       "are given, elasticluster will create new VMs "
+                       "sequentially instead of doing it in parallel.")
         self.add_param('-s', '--storage', metavar="PATH",
                        help="Path to the storage folder. Default: `%s`" %
                             Configurator.default_storage_dir,
@@ -106,6 +108,10 @@ class ElastiCluster(cli.app.CommandLineApp):
         loglevel = max(1, logging.WARNING - 10 * max(0, self.params.verbose))
         log.setLevel(loglevel)
 
+        # In debug mode, avoid forking
+        if self.params.verbose > 3:
+            log.DO_NOT_FORK = True
+
         if not os.path.isdir(self.params.storage):
             # We do not create *all* the parents, but we do create the
             # directory if we can.
