@@ -1,4 +1,3 @@
-#! /usr/bin/env python
 #
 # Copyright (C) 2013 GC3, University of Zurich
 #
@@ -26,12 +25,13 @@ __author__ = 'Riccardo Murri <riccardo.murri@uzh.ch>, ' \
 
 
 # stdlib imports
+import copy
 import httplib2
 import os
 import random
+import threading
 import time
 import uuid
-from multiprocessing import Lock
 
 # External modules
 from apiclient.discovery import build
@@ -51,36 +51,46 @@ from elasticluster.exceptions import ImageError, InstanceError, CloudProviderErr
 #: the OAuth scope for the GCE web API
 GCE_SCOPE = 'https://www.googleapis.com/auth/compute'
 GCE_API_NAME = 'compute'
-GCE_API_VERSION = 'v1beta15'
+GCE_API_VERSION = 'v1'
 GCE_URL = 'https://www.googleapis.com/compute/%s/projects/' % GCE_API_VERSION
 GCE_DEFAULT_ZONE = 'us-central1-a'
 GCE_DEFAULT_SERVICE_EMAIL = 'default'
 GCE_DEFAULT_SCOPES = ['https://www.googleapis.com/auth/devstorage'
-                     '.full_control',
-                  'https://www.googleapis.com/auth/compute']
+                      '.full_control',
+                      'https://www.googleapis.com/auth/compute']
 
 
 class GoogleCloudProvider(AbstractCloudProvider):
+    """Cloud provider for the Google Compute Engine.
+
+    :param str gce_client_id: Client ID to use in OAuth authentication.
+
+    :param str gce_client_secret: Client secret (password) to use in
+                                  OAuth authentication.
+
+    :param str gce_project_id: Project name to log in to GCE.
+
+    :param str zone: gce zone, default is `us-central1-a`
+
+    :param str network: network to use, default is  `default`
+
+    :param str email: service email to use, default is `default`
+
+    :param str storage_path: path to store authentication data (oauth.dat
+                             file). If no path is specified, the login data
+                             has to be entered after every request.
     """
-    Cloud provider for the Google Compute Engine.
-    """
+    __gce_lock = threading.Lock()
 
     def __init__(self, gce_client_id, gce_client_secret, gce_project_id,
                  zone=GCE_DEFAULT_ZONE, network='default',
                  email=GCE_DEFAULT_SERVICE_EMAIL, storage_path=None):
-        """
-        Initialize a provider for the GCE service.
-
-        :param str gce_client_id:     Client ID to use in OAuth authentication.
-        :param str gce_client_secret: Client secret (password) to use in
-         OAuth authentication.
-        :param str gce_project_id:    Project name to log in to GCE.
-        """
         self._client_id = gce_client_id
         self._client_secret = gce_client_secret
         self._project_id = gce_project_id
         self._zone = zone
         self._network = network
+        self._email = email
         self._storage_path = storage_path
 
         # will be initialized upon first connect
@@ -89,17 +99,17 @@ class GoogleCloudProvider(AbstractCloudProvider):
         self._instances = {}
         self._cached_instances = []
         self._images = None
-        self._gce_lock = Lock()
 
     def _connect(self):
-        """
-        Connects to the cloud provider.
+        """Connects to the cloud web services. If this is the first
+        authentication, a web browser will be started to authenticate
+        against google and provide access to elasticluster.
 
-        Also initializes the OAuth credential storage, which might in
-        turn fire up a browser.
+        :return: A Resource object with methods for interacting with the
+                 service.
         """
         # check for existing connection
-        with self._gce_lock:
+        with GoogleCloudProvider.__gce_lock:
             if self._gce:
                 return self._gce
 
@@ -131,20 +141,24 @@ class GoogleCloudProvider(AbstractCloudProvider):
             return self._gce
 
     def _execute_request(self, request):
-        with self._gce_lock:
+        """Helper method to execute a request, since a lock should be used
+        to not fire up multiple requests at the same time.
+
+        :return: Result of `request.execute`
+        """
+        with GoogleCloudProvider.__gce_lock:
             return request.execute(http=self._auth_http)
 
     # The following function was adapted from
     # https://developers.google.com/compute/docs/api/python_guide
     # (function _blocking_call)
     def _wait_until_done(self, response, wait=30):
-        """
-        Blocks until the operation status is done for the given operation.
+        """Blocks until the operation status is done for the given operation.
 
         :param response: The response object used in a previous GCE call.
 
         :param int wait: Wait up to this number of seconds in between
-        successive polling of the GCE status.
+                         successive polling of the GCE status.
         """
 
         gce = self._connect()
@@ -181,10 +195,23 @@ class GoogleCloudProvider(AbstractCloudProvider):
                        username=None,
                        # these params are specific to the
                        # GoogleCloudProvider
-                       instance_name=None):
-        """
-        Starts a new instance with the given properties and returns
+                       instance_name=None, **kwargs):
+        """Starts a new instance with the given properties and returns
         the instance id.
+
+        :param str key_name: name of the ssh key to connect
+        :param str public_key_path: path to ssh public key
+        :param str private_key_path: path to ssh private key
+        :param str security_group: firewall rule definition to apply on the
+                                   instance
+        :param str flavor: machine type to use for the instance
+        :param str image_id: image type (os) to use for the instance
+        :param str image_userdata: command to execute after startup
+        :param str username: username for the given ssh key, default None
+
+        :param str instance_name: name of the instance
+
+        :return: str - instance id of the started instance
         """
         # construct URLs
         project_url = '%s%s' % (GCE_URL, self._project_id)
@@ -198,10 +225,6 @@ class GoogleCloudProvider(AbstractCloudProvider):
 
         # construct the request body
         if instance_name is None:
-            # TODO: it would be nice to have a way to name this
-            # <clustername>-<nodetype>-NNN, e.g.,
-            # "mycluster-compute-001", but we take an easy path to
-            # uniqueness for now.
             instance_name = 'elasticluster-%s' % uuid.uuid4()
 
         public_key_content = file(public_key_path).read()
@@ -209,7 +232,15 @@ class GoogleCloudProvider(AbstractCloudProvider):
         instance = {
             'name': instance_name,
             'machineType': machine_type_url,
-            'image': image_url,
+            'disks': [{
+                'autoDelete': 'true',
+                'boot': 'true',
+                'type': 'PERSISTENT',
+                'initializeParams' : {
+                    'diskName': "%s-disk" % instance_name,
+                    'sourceImage': image_url
+                    }
+                }],
             'networkInterfaces': [
                 {'accessConfigs': [
                     {'type': 'ONE_TO_ONE_NAT',
@@ -218,7 +249,7 @@ class GoogleCloudProvider(AbstractCloudProvider):
                  'network': network_url
                 }],
             'serviceAccounts': [
-                {'email': GCE_DEFAULT_SERVICE_EMAIL,
+                {'email': self._email,
                  'scopes': GCE_DEFAULT_SCOPES
                 }],
             "metadata": {
@@ -245,10 +276,11 @@ class GoogleCloudProvider(AbstractCloudProvider):
             log.error("Error creating instance `%s`" % e)
             raise InstanceError("Error creating instance `%s`" % e)
 
-
     def stop_instance(self, instance_id):
-        """
-        Stops the instance with the given id gracefully.
+        """Stops the instance gracefully.
+
+        :param str instance_id: instance identifier
+        :raises: `InstanceError` if instance can not be stopped
         """
         gce = self._connect()
 
@@ -262,10 +294,10 @@ class GoogleCloudProvider(AbstractCloudProvider):
                                 % (instance_id, e))
 
     def list_instances(self, filter=None):
-        """
-        List instances on GCE, optionally filtering the results.
+        """List instances on GCE, optionally filtering the results.
 
         :param str filter: Filter specification; see https://developers.google.com/compute/docs/reference/latest/instances/list for details.
+        :return: list of instances
         """
         gce = self._connect()
 
@@ -284,11 +316,11 @@ class GoogleCloudProvider(AbstractCloudProvider):
             return list()
 
     def get_ips(self, instance_id):
-        """
-        Fetches the ip addresses (private and public) from the cloud
+        """Retrieves the ip addresses (private and public) from the cloud
         provider by the given instance id.
-        :param instance_id: id of the instance
-        :return: tuple (ip_private, ip_public)
+
+        :param str instance_id: id of the instance
+        :return: list (ips)
         :raises: InstanceError if the ip could not be retrieved.
         """
         gce = self._connect()
@@ -308,21 +340,22 @@ class GoogleCloudProvider(AbstractCloudProvider):
                         ip_public = interfaces[0]['accessConfigs'][0]['natIP']
 
             if ip_private and ip_public:
-                return ip_private, ip_public
+                return [ip_private, ip_public]
             else:
                 raise InstanceError("could not retrieve the ip address for "
                                     "node `%s`, please check the node "
                                     "through the cloud provider interface"
-                                     % instance_id)
+                                    % instance_id)
 
         except (HttpError, CloudProviderError) as e:
             raise InstanceError('could not retrieve the ip address of `%s`: '
                                 '`%s`' % (instance_id, e))
 
     def is_instance_running(self, instance_id):
-        """
-        Return True/False depending on whether the instance with the
-        given id is up and running.
+        """Check whether the instance is up and running.
+
+        :param str instance_id: instance identifier
+        :reutrn: True if instance is running, False otherwise
         """
         items = self.list_instances(filter=('name eq "%s"' % instance_id))
         for item in items:
@@ -331,12 +364,12 @@ class GoogleCloudProvider(AbstractCloudProvider):
         return False
 
     def _get_image_url(self, image_id):
-        """
-        Gets the url for the specified image. Unfortunatly this only works for
-        images uploaded by the user. The images provided by google will not
-        be found.
-        :param image_id: name of the image
-        :return: api url of the image
+        """Gets the url for the specified image. Unfortunatly this only works
+        for images uploaded by the user. The images provided by google will
+        not be found.
+
+        :param str image_id: image identifier
+        :return: str - api url of the image
         """
         gce = self._connect()
         filter = "name eq %s" % image_id
@@ -354,8 +387,8 @@ class GoogleCloudProvider(AbstractCloudProvider):
             raise ImageError("Could not find given image id `%s`" % image_id)
 
     def _check_response(self, response):
-        """
-        Checks the response from GCE for error messages.
+        """Checks the response from GCE for error messages.
+
         :param response: GCE response
         :return: nothing
         :raises: CloudProviderError with error message from GCE
@@ -365,3 +398,17 @@ class GoogleCloudProvider(AbstractCloudProvider):
             raise CloudProviderError("The following error occurred while "
                                      "interacting with the cloud provider "
                                      "`%s`" % error)
+
+    def __getstate__(self):
+        """
+        Overwrites the default dictionary for pickle. Only the gce connection
+        is reset in this method in order to enforce a reconnect.
+        """
+        pickle_dict = copy.deepcopy(self.__dict__)
+
+        # the gce connection might be lost when unpickling, therefore we just
+        # save an empty gce connection to mitigate the problems in the first
+        # place.
+        pickle_dict['_gce'] = None
+
+        return pickle_dict
