@@ -86,9 +86,8 @@ def cluster_summary(cluster):
                   "%s", str(ex))
     msg = """
 Cluster name:     %s
-Cluster template: %s
 Frontend node: %s
-""" % (cluster.name, cluster.template, frontend)
+""" % (cluster.name, frontend)
 
     for cls in cluster.nodes:
         msg += "- %s nodes: %d\n" % (cls, len(cluster.nodes[cls]))
@@ -181,7 +180,11 @@ class Start(AbstractCommand):
                 print("Starting cluster `%s` with %d %s nodes." % (
                     cluster.name, len(cluster.nodes[cls]), cls))
             print("(this may take a while...)")
-            cluster.start()
+            conf = configurator.cluster_conf[cluster_template]
+            min_nodes = dict(
+                (k[:-10], int(v)) for k, v in conf['cluster'].iteritems() if
+                k.endswith('_nodes_min'))
+            cluster.start(min_nodes=min_nodes)
             if self.params.no_setup:
                 print("NOT configuring the cluster as requested.")
             else:
@@ -263,6 +266,8 @@ class ResizeCluster(AbstractCommand):
         parser.add_argument('-r', '--remove', metavar='N1:GROUP1[,N2:GROUP2]',
                             help="Remove N1 nodes of group GROUP1, "
                                  "N2 of group GROUP2 etc...")
+        parser.add_argument('-t', '--template', help='name of the template '
+                                                     'of this cluster')
         parser.add_argument('-v', '--verbose', action='count', default=0,
                             help="Increase verbosity.")
         parser.add_argument('--no-setup', action="store_true", default=False,
@@ -301,6 +306,7 @@ class ResizeCluster(AbstractCommand):
 
         # Get current cluster configuration
         cluster_name = self.params.cluster
+        template = self.params.template
 
         try:
             cluster = configurator.load_cluster(cluster_name)
@@ -312,8 +318,42 @@ class ResizeCluster(AbstractCommand):
         for grp in self.params.nodes_to_add:
             print("Adding %d %s node(s) to the cluster"
                   "" % (self.params.nodes_to_add[grp], grp))
-            for i in range(self.params.nodes_to_add[grp]):
-                cluster.add_node(grp)
+
+            # Currently we can't save which template was used to setup a
+            # cluster, therefore we imply the configuration of the new nodes
+            # to match already existent nodes in this group. If no node was
+            # added to this group yet, it will abort and ask for the
+            # `--template` argument.
+            # TODO: find a better solution for this problem, it makes things
+            #       complicated for the user
+            if (not grp in cluster.nodes or not cluster.nodes[grp]) \
+                    and not template:
+                print "Elasticluster can not infer which template to use for "\
+                      "the new node(s). Please provide the template with " \
+                      "the `-t` or `--template` option"
+                return
+
+            if not template:
+                sample_node = cluster.nodes[grp][0]
+                for i in range(self.params.nodes_to_add[grp]):
+                    cluster.add_node(grp,
+                                     sample_node.image,
+                                     sample_node.image_user,
+                                     sample_node.flavor,
+                                     sample_node.security_group,
+                                     image_userdata=sample_node.image_userdata)
+            else:
+                conf = configurator.cluster_conf[template]
+                conf_kind = conf['nodes'][grp]
+                for i in range(self.params.nodes_to_add[grp]):
+                    image_user = conf['login']['image_user']
+                    userdata = conf_kind.get('image_userdata', '')
+                    cluster.add_node(grp,
+                                     conf_kind['image_id'],
+                                     image_user,
+                                     conf_kind['flavor'],
+                                     conf_kind['security_group'],
+                                     image_userdata=userdata)
 
         for grp in self.params.nodes_to_remove:
             n_to_rm = self.params.nodes_to_remove[grp]
@@ -326,7 +366,7 @@ class ResizeCluster(AbstractCommand):
             if not self.params.yes:
                 # Ask for confirmation.
                 yesno = raw_input(
-                    "Do you want really want to remove them? [yN] ")
+                    "Do you really want to remove them? [yN] ")
                 if yesno.lower() not in ['yes', 'y']:
                     print("Aborting as per user request.")
                     sys.exit(0)
@@ -360,28 +400,20 @@ class ListClusters(AbstractCommand):
     def execute(self):
         configurator = Configurator.fromConfig(
             self.params.config, storage_path=self.params.storage)
-        storage = configurator.create_cluster_storage()
-        cluster_names = storage.get_stored_clusters()
+        repository = configurator.create_repository()
+        clusters = repository.get_all()
 
-        if not cluster_names:
+        if not clusters:
             print("No clusters found.")
         else:
             print("""
 The following clusters have been started.
 Please note that there's no guarantee that they are fully configured:
 """)
-            for name in sorted(cluster_names):
-                try:
-                    cluster = configurator.load_cluster(name)
-                except ConfigurationError, ex:
-                    log.error("gettin information from cluster `%s`: %s",
-                              name, ex)
-                    continue
-                print("%s " % name)
-                print("-" * len(name))
+            for cluster in sorted(clusters):
+                print("%s " % cluster.name)
+                print("-" * len(cluster.name))
                 print("  name:           %s" % cluster.name)
-                print("  template:       %s" % cluster.template)
-                print("  cloud:          %s " % cluster._cloud)
                 for cls in cluster.nodes:
                     print("  - %s nodes: %d" % (cls, len(cluster.nodes[cls])))
                 print("")
@@ -409,21 +441,23 @@ class ListTemplates(AbstractCommand):
             self.params.config, storage_path=self.params.storage)
         config = configurator.cluster_conf
 
-        print("""%d cluster templates found.""" % len(config))
+        print("""%d cluster templates found in configuration file.""" % len(config))
         templates = config.keys()
         for pattern in self.params.clusters:
             templates = [t for t in templates if fnmatch(t, pattern)]
 
+        if self.params.clusters:
+            print("""%d cluter templates found matching pattern(s) '%s'""" % (len(templates), str.join(", ", self.params.clusters)))
+
         for template in templates:
             try:
                 cluster = configurator.create_cluster(template, template)
                 print("""
-name:     %s
-cloud:     %s""" % (template, cluster._cloud))
-                for nodetype in cluster.nodes:
+name:     %s""" % template)
+                for nodekind in cluster.nodes:
                     print("%s nodes: %d" % (
-                        nodetype,
-                        len(cluster.nodes[nodetype])))
+                        nodekind,
+                        len(cluster.nodes[nodekind])))
             except ConfigurationError, ex:
                 log.warning("unable to load cluster `%s`: %s", template, ex)
 
@@ -541,18 +575,25 @@ class SshFrontend(AbstractCommand):
 
         try:
             frontend = cluster.get_frontend_node()
+            # ensure we can connect to the host
+            if not frontend.preferred_ip:
+                # Ensure we can connect to the node, and save the value of `preferred_ip`
+                ssh = frontend.connect()
+                if ssh:
+                    ssh.close()
+                cluster.repository.save_or_update(cluster)
+
         except NodeNotFound, ex:
             log.error("Unable to connect to the frontend node: %s" % str(ex))
             sys.exit(1)
-        host = frontend.ip_public
+        host = frontend.connection_ip()
         username = frontend.image_user
         log.warning("Ignoring known_hosts file.")
         ssh_cmdline = ["ssh",
                        "-i", frontend.user_key_private,
                        "-o", "UserKnownHostsFile=/dev/null",
                        "-o", "StrictHostKeyChecking=no",
-                       '%s@%s' % (username, host),
-        ]
+                       '%s@%s' % (username, host)]
         ssh_cmdline.extend(self.params.ssh_args)
         log.debug("Running command `%s`" % str.join(' ', ssh_cmdline))
         os.execlp("ssh", *ssh_cmdline)
@@ -593,13 +634,12 @@ class SftpFrontend(AbstractCommand):
         except NodeNotFound, ex:
             log.error("Unable to connect to the frontend node: %s" % str(ex))
             sys.exit(1)
-        host = frontend.ip_public
+        host = frontend.connection_ip()
         username = frontend.image_user
         sftp_cmdline = ["sftp",
-                        "-i", frontend.user_key_private,
                         "-o", "UserKnownHostsFile=/dev/null",
                         "-o", "StrictHostKeyChecking=no",
-        ]
+                        "-o", "IdentityFile=%s" % frontend.user_key_private]
         sftp_cmdline.extend(self.params.sftp_args)
         sftp_cmdline.append('%s@%s' % (username, host))
         os.execlp("sftp", *sftp_cmdline)
